import csv
import os.path

import numpy as np
import pandas as pd


COLUMNS = {
    'txNomeParlamentar': 'congressperson_name',
    'idecadastro': 'congressperson_id',
    'nuCarteiraParlamentar': 'congressperson_document',
    'nuLegislatura': 'term',
    'sgUF': 'state',
    'sgPartido': 'party',
    'codLegislatura': 'term_id',
    'numSubCota': 'subquota_number',
    'txtDescricao': 'subquota_description',
    'numEspecificacaoSubCota': 'subquota_group_id',
    'txtDescricaoEspecificacao': 'subquota_group_description',
    'txtFornecedor': 'supplier',
    'txtCNPJCPF': 'cnpj_cpf',
    'txtNumero': 'document_number',
    'indTipoDocumento': 'document_type',
    'datEmissao': 'issue_date',
    'vlrDocumento': 'document_value',
    'vlrGlosa': 'remark_value',
    'vlrLiquido': 'net_value',
    'numMes': 'month',
    'numAno': 'year',
    'numParcela': 'installment',
    'txtPassageiro': 'passenger',
    'txtTrecho': 'leg_of_the_trip',
    'numLote': 'batch_number',
    'numRessarcimento': 'reimbursement_number',
    'vlrRestituicao': 'reimbursement_value',
    'nuDeputadoId': 'applicant_id',
    'ideDocumento': 'document_id',
}
SUBQUOTAS = (
    ('1', 'Maintenance of office supporting parliamentary activity'),
    ('2', 'Locomotion, meal and lodging'),
    ('3', 'Fuels and lubricants'),
    ('4', 'Consultancy, research and technical work'),
    ('5', 'Publicity of parliamentary activity'),
    ('6', 'Purchase of office supplies'),
    ('7', 'Software purchase or renting; Postal services; Subscriptions'),
    ('8', 'Security service provided by specialized company'),
    ('9', 'Flight tickets'),
    ('10', 'Telecommunication'),
    ('11', 'Postal services'),
    ('12', 'Publication subscriptions'),
    ('13', 'Congressperson meal'),
    ('14', 'Lodging, except for congressperson from Distrito Federal'),
    ('15', 'Automotive vehicle renting or watercraft charter'),
    ('119', 'Aircraft renting or charter of aircraft'),
    ('120', 'Automotive vehicle renting or charter'),
    ('121', 'Watercraft renting or charter'),
    ('122', 'Taxi, toll and parking'),
    ('123', 'Terrestrial, maritime and fluvial tickets'),
    ('137', 'Participation in course, talk or similar event'),
    ('999', 'Flight ticket issue')
)
DTYPE = {
    'txNomeParlamentar': np.str,
    'idecadastro': np.str,
    'nuCarteiraParlamentar': np.str,
    'nuLegislatura': np.str,
    'sgUF': np.str,
    'sgPartido': np.str,
    'codLegislatura': np.str,
    'numSubCota': np.str,
    'txtDescricao': np.str,
    'numEspecificacaoSubCota': np.str,
    'txtDescricaoEspecificacao': np.str,
    'txtFornecedor': np.str,
    'txtCNPJCPF': np.str,
    'txtNumero': np.str,
    'indTipoDocumento': np.str,
    'datEmissao': np.str,
    'vlrDocumento': np.float,
    'vlrGlosa': np.str,
    'vlrLiquido': np.float,
    'numMes': np.str,
    'numAno': np.str,
    'numParcela': np.str,
    'txtPassageiro': np.str,
    'txtTrecho': np.str,
    'numLote': np.str,
    'numRessarcimento': np.str,
    'nuDeputadoId': np.str,
    'ideDocumento': np.str,
}


def parse_float(value):
    return float(value.replace(',', '.'))


CONVERTERS = {
    'vlrRestituicao': parse_float,
}
KEY = 'document_id'
AGGREGATED_COLS = {
    'reimbursement_number': 'numbers',
    'net_value': 'total_net_value',
    'reimbursement_value': 'total_value',
}


class ReimbursementsCleaner:
    """
    Perform data cleaning tasks over a reimbursements CSV file.
    """

    def __init__(self, year, path):
        self.year = year
        self.path = path
        self.data = None

    def __call__(self):
        self.load_source_file()
        self.translate()
        self.aggregate_multiple_payments()
        self.save()

    def load_source_file(self):
        file_path = os.path.join(self.path, f'Ano-{self.year}.csv')
        self.data = pd.read_csv(file_path,
                                delimiter=';',
                                quoting=csv.QUOTE_NONE,
                                decimal=',',
                                dtype=DTYPE,
                                low_memory=False)

    def translate(self):
        self.data.rename(columns=COLUMNS, inplace=True)
        for code, name in SUBQUOTAS:
            rows = self.data['subquota_number'] == code
            self.data.loc[rows, 'subquota_description'] = name

    def aggregate_multiple_payments(self):
        self.data = pd.concat([
            self._house_payments(),
            self._non_house_payments(),
        ])

    def save(self):
        file_path = os.path.join(self.path, f'reimbursements-{self.year}.csv')
        self.data.to_csv(file_path, index=False)

    def _house_payments(self):
        data = self.data[self.data['reimbursement_number'] == '0'].copy()
        data.rename(columns=AGGREGATED_COLS, inplace=True)
        data['numbers'] = data['numbers'].apply(lambda val: [val])
        return data

    def _non_house_payments(self):
        data = self.data[self.data['reimbursement_number'] != '0'].copy()
        data.rename(columns=AGGREGATED_COLS, inplace=True)
        attributes = {
            key: 'first' for key in data.columns
            if key is not KEY
        }
        attributes['numbers'] = list
        attributes['total_net_value'] = 'sum'
        attributes['total_value'] = 'sum'
        return data.groupby(KEY, as_index=False).agg(attributes)



import os

from serenata_toolbox import log, settings
from serenata_toolbox.datasets.downloader import Downloader
from serenata_toolbox.datasets.local import LocalDatasets


class Datasets:
    """
    This is a wrapper for three different classes that together handle the
    datasets locally and the download of new ones).

    Datasets class takes one argument: the path to the local directory of the
    dataset files (e.g. data/ or /tmp/serenata-data). The argument is optional
    and the default value is data/ (following the default usage in the main
    repo, serenata-de-amor).

    Inside this object there are two main objects: local and downloader:

    * `Datasets.local` handles listing all local datasets through the property
      `Datasets.local.all` (hint: it's a generator) and deleting local datasets
      with the method `Datasets.local.delete(filename)`;

    * `Datasets.downloader` implements a async manager to download files from
      the remote bucket. It's `Datasets.downloader.download(files)` take the
      path for a single file (str) as argument or an iterable of paths (str).

    :param local_directory: (str) path to local directory of the datasets
    :param timeout: (float) timeout parameter to Downloader,
           None or 0 disables timeout check.
    """

    def __init__(self, local_directory=None, timeout=None):
        if not local_directory:
            local_directory = 'data'

        self.local = LocalDatasets(local_directory)
        self.downloader = Downloader(local_directory, timeout=timeout)


# shortcuts & retrocompatibility

def fetch(filename, destination_path):
    datasets = Datasets(destination_path)
    return datasets.downloader.download(filename)


def fetch_latest_backup(destination_path, force_all=False):
    datasets = Datasets(destination_path)

    if force_all:
        files = datasets.downloader.LATEST
    else:
        files = tuple(
            f for f in datasets.downloader.LATEST
            if not os.path.exists(os.path.join(destination_path, f))
        )

    if not files:
        log.info('You already have all the latest datasets! Nothing to download.')

    return datasets.downloader.download(files)



import os

from serenata_toolbox.datasets.contextmanager import status_message


class LocalDatasets:

    def __init__(self, directory):
        if not all((os.path.exists(directory), os.path.isdir(directory))):
            raise FileNotFoundError('{} is not a directory'.format(directory))

        self.directory = os.path.abspath(directory)

    @property
    def all(self):
        yield from filter(self._is_file, os.listdir(self.directory))

    def delete(self, file_name):
        full_path = os.path.join(self.directory, file_name)
        if not all((os.path.exists(full_path), os.path.isfile(full_path))):
            raise FileNotFoundError('{} is not a flile'.format(full_path))

        with status_message('Deleting {}…'.format(file_name)):
            os.remove(full_path)

    def _is_file(self, filename):
        full_path = os.path.join(self.directory, filename)
        return os.path.isfile(full_path)



import urllib
import xml.etree.ElementTree as ET

import pandas as pd

from serenata_toolbox import log
from serenata_toolbox.datasets.helpers import (
    save_to_csv,
    xml_extract_date,
    xml_extract_datetime,
    xml_extract_text,
)


class SpeechesDataset:

    URL = (
        'http://www.camara.leg.br/SitCamaraWS/SessoesReunioes.asmx/ListarDiscursosPlenario'
        '?dataIni={dataIni}'
        '&dataFim={dataFim}'
        '&codigoSessao=&parteNomeParlamentar=&siglaPartido=&siglaUF='
    )

    def fetch(self, range_start, range_end):
        """
        Fetches speeches from the ListarDiscursosPlenario endpoint of the
        SessoesReunioes (SessionsReunions) API.

        The date range provided should be specified as a string using the
        format supported by the API (%d/%m/%Y)
        """
        range_dates = {'dataIni': range_start, 'dataFim': range_end}
        url = self.URL.format(**range_dates)
        xml = urllib.request.urlopen(url)

        tree = ET.ElementTree(file=xml)
        records = self._parse_speeches(tree.getroot())

        return pd.DataFrame(records, columns=[
            'session_code',
            'session_date',
            'session_num',
            'phase_code',
            'phase_desc',
            'speech_speaker_num',
            'speech_speaker_name',
            'speech_speaker_party',
            'speech_speaker_state',
            'speech_started_at',
            'speech_room_num',
            'speech_insertion_num'
        ])

    @staticmethod
    def _parse_speeches(root):
        for session in root:
            session_code = xml_extract_text(session, 'codigo')
            session_date = xml_extract_date(session, 'data')
            session_num = xml_extract_text(session, 'numero')
            for phase in session.find('fasesSessao'):
                phase_code = xml_extract_text(phase, 'codigo')
                phase_desc = xml_extract_text(phase, 'descricao')
                for speech in phase.find('discursos'):
                    speech_speaker_num = xml_extract_text(speech, 'orador/numero')
                    speech_speaker_name = xml_extract_text(speech, 'orador/nome')
                    speech_speaker_party = xml_extract_text(speech, 'orador/partido')
                    speech_speaker_state = xml_extract_text(speech, 'orador/uf')

                    try:
                        speech_started_at = xml_extract_datetime(speech, 'horaInicioDiscurso')
                    except ValueError as value_error_exception:
                        log.warning("Error parsing speech start time for {} - {}/{} on {}\n{}".format(
                            speech_speaker_name,
                            speech_speaker_party,
                            speech_speaker_state,
                            session_date,
                            value_error_exception))
                        continue

                    speech_room_num = xml_extract_text(speech, 'numeroQuarto')
                    speech_insertion_num = xml_extract_text(speech, 'numeroInsercao')

                    yield [
                        session_code,
                        session_date,
                        session_num,
                        phase_code,
                        phase_desc,
                        speech_speaker_num,
                        speech_speaker_name,
                        speech_speaker_party,
                        speech_speaker_state,
                        speech_started_at,
                        speech_room_num,
                        speech_insertion_num
                    ]


def fetch_speeches(data_dir, range_start, range_end):
    """
    :param data_dir: (str) directory in which the output file will be saved
    :param range_start: (str) date in the format dd/mm/yyyy
    :param range_end: (str) date in the format dd/mm/yyyy
    """
    speeches = SpeechesDataset()
    df = speeches.fetch(range_start, range_end)
    save_to_csv(df, data_dir, "speeches")
    return df



import os.path
from tempfile import gettempdir
from urllib.request import urlretrieve
from zipfile import ZipFile

from .reimbursements_cleaner import ReimbursementsCleaner

URL = 'https://www.camara.leg.br/cotas/Ano-{}.csv.zip'


def extract_zip(zip_path, destination_path):
    zip_file = ZipFile(zip_path, 'r')
    zip_file.extractall(destination_path)
    zip_file.close()


class Reimbursements:
    """
    Get an updated version of the reimbursements dataset for a given year.
    """

    def __init__(self, year, path=None):
        self.year = year
        self.path = path or gettempdir()

    def __call__(self):
        self.fetch()
        self.clean()
        file_path = os.path.join(self.path, f'reimbursements-{self.year}.csv')
        return file_path

    def fetch(self):
        file_path = os.path.join(self.path, f'Ano-{self.year}.zip')
        urlretrieve(URL.format(self.year), file_path)
        extract_zip(file_path, self.path)

    def clean(self):
        ReimbursementsCleaner(self.year, self.path)()



import os.path
from datetime import date
from urllib.error import HTTPError, URLError
from urllib.request import urlretrieve

import pandas as pd

from serenata_toolbox import log


class Dataset:
    URL = 'http://www.senado.gov.br/transparencia/LAI/verba/{}.csv'

    AVAILABLE_YEARS = [year for year in range(2008, date.today().year + 1)]

    def __init__(self, path, years=AVAILABLE_YEARS):
        self.path = path
        self.years = years if isinstance(years, list) else [years]

    def fetch(self):
        retrieved_files = []
        not_found_files = []

        for year in self.years:
            url = self.URL.format(year)
            file_path = os.path.join(self.path, 'federal-senate-{}.csv'.format(year))
            try:
                urlretrieve(url, file_path)
            except HTTPError as http_error_exception:
                log.error('We failed to reach the server')
                log.error('Error code ', http_error_exception.reason)
                log.error("While fetching, Seranata Toolbox didn't find file: {} \n{}".format(
                    file_path,
                    http_error_exception)
                )
                raise http_error_exception
            except URLError as url_error_exception:
                log.error("The server couldn\'t fulfill the request.")
                log.error('Reason: ', url_error_exception.reason)
                log.error("While fetching, Seranata Toolbox didn't find file: {} \n{}".format(
                    file_path,
                    url_error_exception)
                )
                raise url_error_exception
            else:
                retrieved_files.append(file_path)

        return (retrieved_files, not_found_files)

    def translate(self):
        filenames = self._filename_generator('csv')
        not_found_files = []
        translated_files = []

        for filename in filenames:
            csv_path = os.path.join(self.path, filename)
            try:
                self._translate_file(csv_path)
            except FileNotFoundError as file_not_found_error:
                log.error("While translating, Seranata Toolbox didn't find file: {} \n{}".format(
                    csv_path,
                    file_not_found_error)
                )
                raise file_not_found_error
            else:
                translated_files.append(csv_path)

        return (translated_files, not_found_files)

    def clean(self):
        filenames = self._filename_generator('xz')

        merged_dataset = self._merge_files(filenames)

        cleaned_merged_dataset = self._cleanup_dataset(merged_dataset)

        reimbursement_path = os.path.join(self.path, 'federal-senate-reimbursements.xz')
        cleaned_merged_dataset.to_csv(reimbursement_path,
                                      compression='xz',
                                      index=False,
                                      encoding='utf-8')

        return reimbursement_path

    def _filename_generator(self, extension):
        return ['federal-senate-{}.{}'.format(year, extension) for year in self.years]

    @staticmethod
    def _cleanup_dataset(dataset):
        dataset['date'] = pd.to_datetime(dataset['date'], errors='coerce')
        dataset['cnpj_cpf'] = dataset['cnpj_cpf'].str.replace(r'\D', '')

        return dataset

    def _merge_files(self, filenames):
        dataset = pd.DataFrame()

        for filename in filenames:
            file_path = os.path.join(self.path, filename)
            data = pd.read_csv(file_path, encoding='utf-8')
            dataset = pd.concat([dataset, data])

        return dataset

    @staticmethod
    def _translate_file(csv_path):
        output_file_path = csv_path.replace('.csv', '.xz')

        data = pd.read_csv(csv_path,
                           sep=';',
                           encoding="ISO-8859-1",
                           skiprows=1)

        data.columns = [str.lower(column) for column in data.columns]

        data.rename(columns={
            'ano': 'year',
            'mes': 'month',
            'senador': 'congressperson_name',
            'tipo_despesa': 'expense_type',
            'cnpj_cpf': 'cnpj_cpf',
            'fornecedor': 'supplier',
            'documento': 'document_id',
            'data': 'date',
            'detalhamento': 'expense_details',
            'valor_reembolsado': 'reimbursement_value',
        }, inplace=True)

        data['expense_type'] = data['expense_type'].astype('category')

        data['expense_type'] = \
            data['expense_type'].astype('category')

        pt_categories = (
            'Aluguel de imóveis para escritório político, compreendendo despesas concernentes a eles.',
            ('Aquisição de material de consumo para uso no escritório político, inclusive aquisição ou locação'
                ' de software, despesas postais, aquisição de publicações, locação de móveis e de equipamentos. '),
            ('Contratação de consultorias, assessorias, pesquisas, trabalhos técnicos e outros serviços de '
                'apoio ao exercício do mandato parlamentar'),
            'Divulgação da atividade parlamentar',
            'Locomoção, hospedagem, alimentação, combustíveis e lubrificantes',
            'Passagens aéreas, aquáticas e terrestres nacionais',
            'Serviços de Segurança Privada'
        )
        en_categories = (
            'Rent of real estate for political office, comprising expenses concerning them',
            ('Acquisition of consumables for use in the political office, including acquisition or leasing of'
                ' software, postal expenses, acquisition of publications, rental of furniture and equipment'),
            ('Recruitment of consultancies, advisory services, research, technical work and other services'
                ' in support of the exercise of the parliamentary mandate'),
            'Publicity of parliamentary activity',
            'Locomotion, lodging, food, fuels and lubricants',
            'National air, water and land transport',
            'Private Security Services'
        )
        categories = dict(zip(pt_categories, en_categories))

        categories = [categories[cat] for cat in data['expense_type'].cat.categories]

        data['expense_type'].cat.rename_categories(categories, inplace=True)

        data.to_csv(output_file_path, compression='xz', index=False, encoding='utf-8')

        return output_file_path



import urllib
import xml.etree.ElementTree as ET

import pandas as pd

from serenata_toolbox import log
from serenata_toolbox.datasets.helpers import (
    save_to_csv,
    translate_column,
    xml_extract_text,
)


class DeputiesDataset:

    URL = 'http://www.camara.leg.br/SitCamaraWS/deputados.asmx/ObterDeputados'

    def fetch(self):
        """
        Fetches the list of deputies for the current term.
        """
        xml = urllib.request.urlopen(self.URL)

        tree = ET.ElementTree(file=xml)
        records = self._parse_deputies(tree.getroot())

        df = pd.DataFrame(records, columns=(
            'congressperson_id',
            'budget_id',
            'condition',
            'congressperson_document',
            'civil_name',
            'congressperson_name',
            'picture_url',
            'gender',
            'state',
            'party',
            'phone_number',
            'email'
        ))
        return self._translate(df)

    @staticmethod
    def _parse_deputies(root):
        for deputy in root:
            yield (
                xml_extract_text(deputy, 'ideCadastro'),
                xml_extract_text(deputy, 'codOrcamento'),
                xml_extract_text(deputy, 'condicao'),
                xml_extract_text(deputy, 'matricula'),
                xml_extract_text(deputy, 'nome'),
                xml_extract_text(deputy, 'nomeParlamentar'),
                xml_extract_text(deputy, 'urlFoto'),
                xml_extract_text(deputy, 'sexo'),
                xml_extract_text(deputy, 'uf'),
                xml_extract_text(deputy, 'partido'),
                xml_extract_text(deputy, 'fone'),
                xml_extract_text(deputy, 'email'),
            )

    @staticmethod
    def _translate(df):
        translate_column(df, 'gender', {
            'masculino': 'male',
            'feminino': 'female',
        })

        translate_column(df, 'condition', {
            'Titular': 'Holder',
            'Suplente': 'Substitute',
        })

        return df


def fetch_deputies(data_dir):
    """
    :param data_dir: (str) directory in which the output file will be saved
    """
    deputies = DeputiesDataset()
    df = deputies.fetch()
    save_to_csv(df, data_dir, "deputies")

    holders = df.condition == 'Holder'
    substitutes = df.condition == 'Substitute'
    log.info("Total deputies:", len(df))
    log.info("Holder deputies:", len(df[holders]))
    log.info("Substitute deputies:", len(df[substitutes]))
    return df
